from itertools import chain, combinations
from typing import Union

import numpy as np
import torch
import torch.distributions as dist
from pythae.models.base.base_utils import ModelOutput

from multivae.data.datasets.base import IncompleteDataset, MultimodalBaseDataset
from multivae.models.nn.default_architectures import (
    BaseDictDecodersMultiLatents,
    BaseDictEncoders_MultiLatents,
)

from ..base import BaseMultiVAE
from .mopoe_config import MoPoEConfig


class MoPoE(BaseMultiVAE):

    """

    Implementation for the Mixture of Product of experts model from
    'Generalized Multimodal ELBO' Sutter 2021 (https://arxiv.org/abs/2105.02470)

    This implementation is heavily based on the official one at
    https://github.com/thomassutter/MoPoE

    """

    def __init__(
        self, model_config: MoPoEConfig, encoders: dict = None, decoders: dict = None
    ):
        super().__init__(model_config, encoders, decoders)

        self.beta = model_config.beta
        self.multiple_latent_spaces = model_config.use_modality_specific_spaces
        self.model_name = "MoPoE"

        list_subsets = self.model_config.subsets
        if type(list_subsets) == dict:
            list_subsets = list(list_subsets.values())
        if list_subsets is None:
            list_subsets = self.all_subsets()
        self.set_subsets(list_subsets)

        if model_config.use_modality_specific_spaces:
            self.style_dims = model_config.modalities_specific_dim
            if model_config.modalities_specific_dim is None:
                raise AttributeError(
                    "Please provide dimensions for the modalities"
                    "specific latent spaces"
                )
            if encoders is None:
                encoders = BaseDictEncoders_MultiLatents(
                    input_dims=model_config.input_dims,
                    latent_dim=model_config.latent_dim,
                    modality_dims=model_config.modalities_specific_dim,
                )
                self.set_encoders(encoders)

            if decoders is None:
                decoders = BaseDictDecodersMultiLatents(
                    input_dims=model_config.input_dims,
                    latent_dim=model_config.latent_dim,
                    modality_dims=model_config.modalities_specific_dim,
                )
                self.set_decoders(decoders)

    def all_subsets(self):
        """
        Returns a list containing all possible subsets of the modalities.
        (But the empty one)
        """
        xs = list(self.encoders.keys())
        # note we return an iterator rather than a list
        subsets_list = chain.from_iterable(
            combinations(xs, n) for n in range(len(xs) + 1)
        )
        return subsets_list

    def set_subsets(self, subsets_list):
        """
        Builds a dictionary of the subsets.
        The keys are the subset_names created by concatenating the modalities' names.
        The values are the list of modalities names.
        """
        subsets = dict()
        for k, mod_names in enumerate(subsets_list):
            mods = []
            for l, mod_name in enumerate(sorted(mod_names)):
                if (mod_name not in self.encoders.keys()) and (mod_name != ""):
                    raise AttributeError(
                        f"The provided subsets list contains unknown modality name {mod_name}."
                        " that is not the encoders dictionary or inputs_dim dictionary."
                    )
                mods.append(mod_name)
            key = "_".join(sorted(mod_names))
            subsets[key] = mods
        self.subsets = subsets
        self.model_config.subsets = subsets
        return

    def reparameterize(self, mu, logvar):
        std = logvar.mul(0.5).exp_()
        z = dist.Normal(mu, std).rsample()
        return z

    def calc_joint_divergence(
        self, mus: torch.Tensor, logvars: torch.Tensor, weights: torch.Tensor
    ):
        """Computes the KL divergence between the mixture of experts and the prior, by
        developping into the sum of the tractable KLs divergences of each expert.

        Args:
            mus (Tensor): The means of the experts. (n_subset,n_samples, latent_dim)
            logvars (Tensor): The logvars of the experts.(n_subset,n_samples, latent_dim)
            weights (Tensor): The weights of the experts.(n_subset,n_samples)


        Returns:
            Tensor, Tensor: The group divergence summed over modalities, A tensor containing the KL terms for each experts.
        """
        weights = weights.clone()

        num_mods = mus.shape[0]
        num_samples = mus.shape[1]
        klds = torch.zeros(num_mods, num_samples)

        device = mus.device
        klds = klds.to(device)
        weights = weights.to(device)
        for k in range(0, num_mods):
            kld_ind = -0.5 * (
                1 - logvars[k, :, :].exp() - mus[k, :, :].pow(2) + logvars[k, :, :]
            ).sum(-1)

            klds[k, :] = kld_ind

        group_div = (
            (weights * klds).sum(dim=0).mean()
        )  # sum over experts, mean over samples

        divs = dict()
        divs["joint_divergence"] = group_div
        return divs

    def forward(self, inputs: MultimodalBaseDataset, **kwargs) -> ModelOutput:
        # Compute latents parameters for all subsets
        latents = self.inference(inputs)
        results = dict()

        # Compute the divergence to the prior
        shared_embeddings = self.reparameterize(
            latents["joint"][0], latents["joint"][1]
        )

        div = self.calc_joint_divergence(
            latents["mus"], latents["logvars"], latents["weights"]
        )
        for k, key in enumerate(div.keys()):
            results[key] = div[key]

        # Compute the reconstruction losses for each modality
        loss = 0
        kld = results["joint_divergence"]
        for m_key in self.encoders.keys():
            # reconstruct this modality from the shared embeddings representation

            if self.multiple_latent_spaces:
                try:
                    style_embeddings = latents["modalities"][m_key].style_embedding
                    full_embedding = torch.cat(
                        [shared_embeddings, style_embeddings], dim=-1
                    )
                except:
                    raise AttributeError(
                        " model_config.use_modality_specific_spaces is True "
                        f"but encoder output for modality {m_key} doesn't have a "
                        "style_embedding attribute. "
                        "When using multiple latent spaces, the encoders' output"
                        "should be of the form : ModelOuput(embedding = ...,"
                        "style_embedding = ...,log_covariance = ..., style_log_covariance = ...)"
                    )
            else:
                full_embedding = shared_embeddings

            recon = self.decoders[m_key](full_embedding).reconstruction
            m_rec = (
                (
                    -self.recon_log_probs[m_key](recon, inputs.data[m_key])
                    * self.rescale_factors[m_key]
                )
                .view(recon.size(0), -1)
                .sum(-1)
            )

            # reconstruction loss
            if hasattr(inputs, "masks"):
                results["recon_" + m_key] = (m_rec * inputs.masks[m_key].float()).mean()
            else:
                results["recon_" + m_key] = m_rec.mean()

            loss += results["recon_" + m_key]

            # If using modality specific latent spaces, add modality specific klds
            if self.multiple_latent_spaces:
                style_mu = latents["modalities"][m_key].style_embedding
                style_log_var = latents["modalities"][m_key].style_log_covariance
                style_kld = -0.5 * (
                    1 - style_log_var.exp() - style_mu.pow(2) + style_log_var
                ).view(style_mu.size(0), -1).sum(-1)

                if hasattr(inputs, "masks"):
                    style_kld *= inputs.masks[m_key].float()

                kld += style_kld.mean() * self.model_config.beta_style

        loss = loss + self.beta * kld

        return ModelOutput(loss=loss, metrics=results)

    def modality_encode(
        self, inputs: Union[MultimodalBaseDataset, IncompleteDataset], **kwargs
    ):
        """

        Computes for each modality, the parameters mu and logvar of the
        unimodal posterior.

        Args:
            inputs (MultimodalBaseDataset): The data to encode.

        Returns:
            dict: Containing for each modality the encoder output.
        """
        encoders_outputs = dict()
        for m, m_key in enumerate(self.encoders.keys()):
            input_modality = inputs.data[m_key]
            output = self.encoders[m_key](input_modality)
            encoders_outputs[m_key] = output

        return encoders_outputs

    def poe(self, mu, logvar, eps=1e-8):
        var = torch.exp(logvar) + eps
        # precision of i-th Gaussian expert at point x
        T = 1.0 / var
        pd_mu = torch.sum(mu * T, dim=0) / torch.sum(T, dim=0)
        pd_var = 1.0 / torch.sum(T, dim=0)
        pd_logvar = torch.log(pd_var)
        return pd_mu, pd_logvar

    def poe_fusion(self, mus: torch.Tensor, logvars: torch.Tensor, weights=None):
        # Following the original implementation : add the prior when we consider the
        # subset that contains all the modalities
        if mus.shape[0] == len(self.encoders.keys()):
            num_samples = mus[0].shape[0]
            device = mus.device
            mus = torch.cat(
                (mus, torch.zeros(1, num_samples, self.latent_dim).to(device)), dim=0
            )
            logvars = torch.cat(
                (logvars, torch.zeros(1, num_samples, self.latent_dim).to(device)),
                dim=0,
            )
        mu_poe, logvar_poe = self.poe(mus, logvars)
        return [mu_poe, logvar_poe]

    def subset_mask(self, inputs: IncompleteDataset, subset: Union[list, tuple]):
        """
        Returns a filter of the samples available in all the modalities contained in subset.
        """

        filter = torch.tensor(
            True,
        ).to(inputs.masks[subset[0]].device)
        for mod in subset:
            filter = torch.logical_and(filter, inputs.masks[mod])

        return filter

    def inference(self, inputs: MultimodalBaseDataset, **kwargs):
        """

        Args:
            inputs (MultimodalBaseDataset): The data.

        Returns:
            dict: all the subset and joint posteriors parameters.
        """

        latents = dict()
        enc_mods = self.modality_encode(inputs)
        latents["modalities"] = enc_mods
        device = inputs.data[list(inputs.data.keys())[0]].device

        mus = torch.Tensor().to(device)
        logvars = torch.Tensor().to(device)
        distr_subsets = dict()
        availabilities = []

        for k, s_key in enumerate(self.subsets.keys()):
            if s_key != "":
                mods = self.subsets[s_key]
                mus_subset = torch.Tensor().to(device)
                logvars_subset = torch.Tensor().to(device)

                if hasattr(inputs, "masks"):
                    filter = self.subset_mask(inputs, mods)
                    availabilities.append(filter)

                for m, mod in enumerate(mods):
                    mus_mod = enc_mods[mod].embedding
                    log_vars_mod = enc_mods[mod].log_covariance

                    mus_subset = torch.cat((mus_subset, mus_mod.unsqueeze(0)), dim=0)

                    logvars_subset = torch.cat(
                        (logvars_subset, log_vars_mod.unsqueeze(0)), dim=0
                    )
                # Case with only one sample : adapt the shape
                if len(mus_subset.shape) == 2:
                    mus_subset = mus_subset.unsqueeze(1)
                    logvars_subset = logvars_subset.unsqueeze(1)

                s_mu, s_logvar = self.poe_fusion(mus_subset, logvars_subset)

                distr_subsets[s_key] = [s_mu, s_logvar]

                # Add the subset posterior to be part of the mixture of experts
                mus = torch.cat((mus, s_mu.unsqueeze(0)), dim=0)
                logvars = torch.cat((logvars, s_logvar.unsqueeze(0)), dim=0)

        if hasattr(inputs, "masks"):
            # if we have an incomplete dataset, we need to randomly choose
            # from the mixture of available experts
            availabilities = torch.stack(availabilities, dim=0).float()
            if len(availabilities.shape) == 1:
                availabilities = availabilities.unsqueeze(1)
            availabilities /= torch.sum(availabilities, dim=0)  # (n_subset,n_samples)

            joint_mu, joint_logvar = self.random_mixture_component_selection(
                mus, logvars, availabilities
            )
            weights = availabilities
        else:
            weights = (1 / float(mus.shape[0])) * torch.ones(mus.shape[0]).to(device)
            joint_mu, joint_logvar = self.deterministic_mixture_component_selection(
                mus, logvars, weights
            )
            weights = (1 / float(mus.shape[0])) * torch.ones(
                mus.shape[0], mus.shape[1]
            ).to(device)

        latents["mus"] = mus
        latents["logvars"] = logvars
        latents["weights"] = weights
        latents["joint"] = [joint_mu, joint_logvar]
        latents["subsets"] = distr_subsets
        return latents

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ) -> ModelOutput:
        cond_mod = super().encode(inputs, cond_mod, N, **kwargs).cond_mod

        # Compute the str associated to the subset
        key = "_".join(sorted(cond_mod))
        return_mean = kwargs.pop("return_mean", False)

        # If the dataset is incomplete, keep only the samples availables in all cond_mod
        # modalities

        latents_subsets = self.inference(inputs)

        mu, log_var = latents_subsets["subsets"][key]

        sample_shape = [N] if N > 1 else []
        if return_mean:
            if len(cond_mod) == self.n_modalities:
                # joint posterior mean
                mu = torch.stack(
                    [
                        latents_subsets["subsets"][k][0]
                        for k in latents_subsets["subsets"]
                    ]
                ).mean(0)

            z = torch.stack([mu] * N) if N > 1 else mu
        else:
            z = dist.Normal(mu, torch.exp(0.5 * log_var)).rsample(sample_shape)

        flatten = kwargs.pop("flatten", False)
        if flatten:
            z = z.reshape(-1, self.latent_dim)

        if self.multiple_latent_spaces:
            modalities_z = dict()
            for m in self.encoders:
                if m in cond_mod:
                    mu_style = latents_subsets["modalities"][m].style_embedding
                    log_var_style = latents_subsets["modalities"][
                        m
                    ].style_log_covariance
                else:
                    mu_style = torch.zeros((len(mu), self.style_dims[m])).to(mu.device)
                    log_var_style = torch.zeros((len(mu), self.style_dims[m])).to(
                        mu.device
                    )
                modalities_z[m] = dist.Normal(
                    mu_style, torch.exp(0.5 * log_var_style)
                ).rsample(sample_shape)
                if flatten:
                    modalities_z[m] = modalities_z[m].reshape(-1, self.style_dims[m])

            return ModelOutput(z=z, one_latent_space=False, modalities_z=modalities_z)
        else:
            return ModelOutput(z=z, one_latent_space=True)

    def random_mixture_component_selection(self, mus, logvars, availabilities):
        """
        Randomly select a subset for each sample among the available subsets.

        Args:
            mus (tensor): (n_subset,n_samples,latent_dim) the means of subset posterior.
            logvars (tensor): (n_subset,n_samples,latent_dim) the log covariance of subset posterior.
            availabilities (tensor): (n_subset,n_samples) boolean tensor.
        """

        probs = availabilities.permute(1, 0)  # n_samples,n_subset
        choice = dist.OneHotCategorical(probs=probs).sample()

        mus_ = mus.permute(1, 0, 2)  # n_samples, n_subset,latent_dim
        logvars_ = logvars.permute(1, 0, 2)

        mus_ = mus_[choice.bool()]
        logvars_ = logvars_[choice.bool()]
        return mus_, logvars_

    def deterministic_mixture_component_selection(self, mus, logvars, w_modalities):
        """
        Associate a subset mu and log_covariance per sample in a balanced way, so that the proportion
        of samples per subset correspond to w_modalities.
        """

        num_components = mus.shape[0]  # number of components
        num_samples = mus.shape[1]

        idx_start = []
        idx_end = []
        for k in range(0, num_components):
            if k == 0:
                i_start = 0
            else:
                i_start = int(idx_end[k - 1])
            if k == w_modalities.shape[0] - 1:
                i_end = num_samples
            else:
                i_end = i_start + int(torch.floor(num_samples * w_modalities[k]))
            idx_start.append(i_start)
            idx_end.append(i_end)
        idx_end[-1] = num_samples
        mu_sel = torch.cat(
            [mus[k, idx_start[k] : idx_end[k], :] for k in range(w_modalities.shape[0])]
        )
        logvar_sel = torch.cat(
            [
                logvars[k, idx_start[k] : idx_end[k], :]
                for k in range(w_modalities.shape[0])
            ]
        )
        return [mu_sel, logvar_sel]

    def compute_joint_nll(
        self,
        inputs: Union[MultimodalBaseDataset, IncompleteDataset],
        K: int = 1000,
        batch_size_K: int = 100,
    ):
        """
        Computes the joint negative log-likelihood using the MoPoE posterior as importance sampling distribution.
        The result is summed over the input batch.

        In the original code, the product of experts is used as inference distribution
        for computing the nll instead of the MoPoe, but that is less coherent with the definition of the
        MoPoE definition as the joint posterior.
        """

        self.eval()

        # Compute the parameters of the joint posterior
        infer = self.inference(inputs)
        mu, log_var = infer["joint"]
        mus_subset = infer["mus"]
        log_vars_subset = infer["logvars"]

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        # And sample from the posterior
        z_joint = qz_xy.rsample([K])  # shape K x n_data x latent_dim
        z_joint = z_joint.permute(1, 0, 2)
        n_data, _, latent_dim = z_joint.shape

        # Then iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0
        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)
            lnpxs = []
            while start_idx < stop_idx:
                latents = z_joint[i][start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in inputs.data:
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, nb_channels, w, h)
                    x_m = inputs.data[mod][i]  # (nb_channels, w, h)

                    lpx_zs += (
                        self.recon_log_probs[mod](recon, x_m)
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = dist.Normal(0, 1)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y) = -ln (1/S \sum q(z|x_s))

                qz_xs = [
                    dist.Normal(
                        mus_subset[j][i], torch.exp(0.5 * log_vars_subset[j][i])
                    )
                    for j in range(len(mus_subset))
                ]
                lqz_xs = torch.stack([q.log_prob(latents).sum(-1) for q in qz_xs])
                lqz_xy = torch.logsumexp(lqz_xs, dim=0) - np.log(
                    len(lqz_xs)
                )  # log_mean_exp

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(stop_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll

    @torch.no_grad()
    def compute_joint_nll_from_subset_encoding(
        self,
        subset,
        inputs: Union[MultimodalBaseDataset, IncompleteDataset],
        K: int = 1000,
        batch_size_K: int = 100,
    ):
        """
        Computes the joint negative log-likelihood using the PoE posterior as importance sampling distribution.
        The result is summed over the input batch.
        """
        self.eval()

        # Only keep the samples complete with regard to the subset modalities
        if hasattr(inputs, "masks"):
            filter = self.subset_mask(inputs, subset)
            filtered_inputs = MultimodalBaseDataset(
                data={m: inputs.data[m][filter] for m in inputs.data}
            )

        else:
            filtered_inputs = inputs
        subset_name = "_".join(sorted(subset))
        # Compute the parameters of the joint posterior
        mu, log_var = self.inference(filtered_inputs)["subsets"][subset_name]

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        # And sample from the posterior
        z_joint = qz_xy.rsample([K])  # shape K x n_data x latent_dim
        z_joint = z_joint.permute(1, 0, 2)
        n_data, _, latent_dim = z_joint.shape

        # Then iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0
        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)

            lnpxs = []
            while start_idx < stop_idx:
                latents = z_joint[i][start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in inputs.data:
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, nb_channels, w, h)
                    x_m = inputs.data[mod][i]  # (nb_channels, w, h)

                    lpx_zs += (
                        self.recon_log_probs[mod](recon, x_m)
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = dist.Normal(0, 1)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y))
                qz_xy = dist.Normal(mu[i], sigma[i])
                lqz_xy = qz_xy.log_prob(latents).sum(dim=-1)

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(stop_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll

    def compute_joint_nll_paper(
        self,
        inputs: Union[MultimodalBaseDataset, IncompleteDataset],
        K: int = 1000,
        batch_size_K: int = 100,
    ):
        entire_subset = list(self.encoders.keys())
        return self.compute_joint_nll_from_subset_encoding(
            entire_subset, inputs, K, batch_size_K
        )



import os
from dataclasses import field
from typing import Union

import torch.nn as nn
from pydantic.dataclasses import dataclass
from pythae.config import BaseConfig


@dataclass
class BaseTrainerConfig(BaseConfig):
    """
    BaseTrainer config class stating the main training arguments.

    Parameters:

        output_dir (str): The directory where model checkpoints, configs and final
            model will be stored. Default: None.
        per_device_train_batch_size (int): The number of training samples per batch and per device.
            Default 64
        per_device_eval_batch_size (int): The number of evaluation samples per batch and per device.
            Default 64
        num_epochs (int): The maximal number of epochs for training. Default: 100
        train_dataloader_num_workers (int): Number of subprocesses to use for train data loading.
            0 means that the data will be loaded in the main process. Default: 0
        eval_dataloader_num_workers (int): Number of subprocesses to use for evaluation data
            loading. 0 means that the data will be loaded in the main process. Default: 0
        optimizer_cls (str): The name of the `torch.optim.Optimizer` used for
            training. Default: :class:`~torch.optim.Adam`.
        optimizer_params (dict): A dict containing the parameters to use for the
            `torch.optim.Optimizer`. If None, uses the default parameters. Default: None.
        scheduler_cls (str): The name of the `torch.optim.lr_scheduler` used for
            training. If None, no scheduler is used. Default None.
        scheduler_params (dict): A dict containing the parameters to use for the
            `torch.optim.le_scheduler`. If None, uses the default parameters. Default: None.
        learning_rate (int): The learning rate applied to the `Optimizer`. Default: 1e-4
        steps_saving (int): A model checkpoint will be saved every `steps_saving` epoch.
            Default: None
        steps_predict (int): A prediction using the best model will be run every `steps_predict`
            epoch. Default: None
        keep_best_on_train (bool): Whether to keep the best model on the train set. Default: False
        seed (int): The random seed for reproducibility
        no_cuda (bool): Disable `cuda` training. Default: False
        world_size (int): The total number of process to run. Default: -1
        local_rank (int): The rank of the node for distributed training. Default: -1
        rank (int): The rank of the process for distributed training. Default: -1
        dist_backend (str): The distributed backend to use. Default: 'nccl'
        master_addr (str): The master address for distributed training. Default: 'localhost'
        master_port (str): The master port for distributed training. Default: '12345'
        start_keep_best (int): The epoch to start keeping track of the best loss. Before that the model
            is automatically saved.
    """

    output_dir: str = None
    per_device_train_batch_size: int = 64
    per_device_eval_batch_size: int = 64
    num_epochs: int = 100
    train_dataloader_num_workers: int = 0
    eval_dataloader_num_workers: int = 0
    optimizer_cls: str = "Adam"
    optimizer_params: Union[dict, None] = None
    scheduler_cls: Union[str, None] = None
    scheduler_params: Union[dict, None] = None
    learning_rate: float = 1e-4
    steps_saving: Union[int, None] = None
    steps_predict: Union[int, None] = None
    keep_best_on_train: bool = False
    seed: int = 8
    no_cuda: bool = False
    world_size: int = field(default=-1)
    local_rank: int = field(default=-1)
    rank: int = field(default=-1)
    dist_backend: str = field(default="nccl")
    master_addr: str = field(default="localhost")
    master_port: str = field(default="12345")
    start_keep_best_epoch: int = 0
    drop_last: bool = False

    def __post_init__(self):
        super().__post_init__()
        env_local_rank = int(os.environ.get("LOCAL_RANK", -1))
        if self.local_rank == -1 and env_local_rank != -1:
            self.local_rank = env_local_rank

        env_world_size = int(os.environ.get("WORLD_SIZE", -1))
        if self.world_size == -1 and env_world_size != -1:
            self.world_size = env_world_size

        env_rank = int(os.environ.get("RANK", -1))
        if self.rank == -1 and env_rank != -1:
            self.rank = env_rank

        env_master_addr = os.environ.get("MASTER_ADDR", "localhost")
        if self.master_addr == "localhost" and env_master_addr != "localhost":
            self.master_addr = env_master_addr
        os.environ["MASTER_ADDR"] = self.master_addr

        env_master_port = os.environ.get("MASTER_PORT", "12345")
        if self.master_port == "12345" and env_master_port != "12345":
            self.master_port = env_master_port
        os.environ["MASTER_PORT"] = self.master_port

    def __post_init_post_parse__(self):
        """Check compatibilty"""
        try:
            import torch.optim as optim

            optimizer_cls = getattr(optim, self.optimizer_cls)
        except AttributeError as e:
            raise AttributeError(
                f"Unable to import `{self.optimizer_cls}` optimizer from 'torch.optim'. "
                "Check spelling and that it is part of 'torch.optim.Optimizers.'"
            )
        if self.optimizer_params is not None:
            try:
                optimizer = optimizer_cls(
                    nn.Linear(2, 2).parameters(),
                    lr=self.learning_rate,
                    **self.optimizer_params,
                )
            except TypeError as e:
                raise TypeError(
                    "Error in optimizer's parameters. Check that the provided dict contains only "
                    f"keys and values suitable for `{optimizer_cls}` optimizer. "
                    f"Got {self.optimizer_params} as parameters.\n"
                    f"Exception raised: {type(e)} with message: " + str(e)
                ) from e
        else:
            optimizer = optimizer_cls(
                nn.Linear(2, 2).parameters(), lr=self.learning_rate
            )

        if self.scheduler_cls is not None:
            try:
                import torch.optim.lr_scheduler as schedulers

                scheduder_cls = getattr(schedulers, self.scheduler_cls)
            except AttributeError as e:
                raise AttributeError(
                    f"Unable to import `{self.scheduler_cls}` scheduler from "
                    "'torch.optim.lr_scheduler'. Check spelling and that it is part of "
                    "'torch.optim.lr_scheduler.'"
                )

            if self.scheduler_params is not None:
                try:
                    scheduder_cls(optimizer, **self.scheduler_params)
                except TypeError as e:
                    raise TypeError(
                        "Error in scheduler's parameters. Check that the provided dict contains only "
                        f"keys and values suitable for `{scheduder_cls}` scheduler. "
                        f"Got {self.scheduler_params} as parameters.\n"
                        f"Exception raised: {type(e)} with message: " + str(e)
                    ) from e



import datetime
import logging
from pathlib import Path

import torch
from torch.utils.data import DataLoader

from multivae.data import MultimodalBaseDataset
from multivae.models.base import BaseMultiVAE
from multivae.samplers.base import BaseSampler

from .evaluator_config import EvaluatorConfig


class Evaluator:
    """
    Base class for computing metrics.

    Args:
        model (BaseMultiVAE) : The model to evaluate.
        test_dataset (MultimodalBaseDataset) : The dataset to use for computing the metrics.
        output (str) : The folder path to save metrics. The metrics will be saved in a metrics.txt file.
        eval_config (EvaluatorConfig) : The configuration class to specify parameters for the evaluation.
        sampler (BaseSampler) : A custom sampler for sampling from the common latent space. If None is provided, samples
            are generated from the prior.


    """

    def __init__(
        self,
        model: BaseMultiVAE,
        test_dataset: MultimodalBaseDataset,
        output: str = None,
        eval_config=EvaluatorConfig(),
        sampler: BaseSampler = None,
    ) -> None:
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = model.to(self.device).eval()
        self.n_data = len(test_dataset)
        self.batch_size = eval_config.batch_size
        self.test_dataset = test_dataset
        self.test_loader = DataLoader(test_dataset, batch_size=eval_config.batch_size)
        if output is not None:
            Path(output).mkdir(parents=True, exist_ok=True)

        self.output = output

        self.set_logger(output)
        self.set_wandb(eval_config.wandb_path)
        self.metrics = {}
        self.sampler = sampler
        if self.sampler is not None:
            if not sampler.is_fitted:
                raise AttributeError(
                    "The provided sampler is not fitted."
                    "Please fit the sampler before using it in the evaluator module."
                )

    def set_logger(self, output):
        evaluator_id = (
            str(datetime.datetime.now())[0:19].replace(" ", "_").replace(":", "-")
        )
        logger = logging.getLogger(evaluator_id)
        logger.setLevel(logging.INFO)

        # our first handler is a console handler
        self.console_handler = logging.StreamHandler()
        logger.addHandler(self.console_handler)

        # the second handler is a file handler
        if output is not None:
            self.file_handler = logging.FileHandler(output + "/metrics.log")
            logger.addHandler(self.file_handler)

        self.logger = logger

    def set_wandb(self, wandb_path):
        if wandb_path is None:
            self.wandb_run = None
            return
        else:
            entity, project, run_id = tuple(wandb_path.split("/"))
            try:
                import wandb
            except:
                raise ModuleNotFoundError(
                    "You provided a wandb_path, but `wandb` package is not installed. Run `pip install wandb`"
                )

            self.wandb_run = wandb.init(
                entity=entity, project=project, id=run_id, resume="must", reinit=True
            )
            return

    def log_to_wandb(self): # pragma: no cover
        if self.wandb_run is not None:
            self.wandb_run.log(self.metrics)

    def finish(self):
        """Removes handlers and finish the wandb run."""

        self.logger.removeHandler(self.console_handler)
        if hasattr(self, "file_handler"):
            self.logger.removeHandler(self.file_handler)

        if self.wandb_run is not None:
            self.wandb_run.finish()



import os
import warnings

import numpy as np
import pytest
import torch
from pythae.data.datasets import DatasetOutput

from multivae.data.datasets.base import MultimodalBaseDataset
from multivae.data.datasets.mmnist import MMNISTDataset


class Test:
    @pytest.fixture(params=[0.2, 0])
    def input_dataset_test(self, request):
        data_path = "~/scratch/data"
        split = "test"
        missing_ratio = request.param

        return dict(data_path=data_path, split=split, missing_ratio=missing_ratio)

    def test_create_dataset(self, input_dataset_test):
        if os.path.exists(
            os.path.join(input_dataset_test["data_path"], "MMNIST", "test", "m0.pt")
        ):
            dataset = MMNISTDataset(**input_dataset_test)
            assert isinstance(dataset, MultimodalBaseDataset)
            sample = dataset[0]
            assert type(sample) == DatasetOutput
            assert isinstance(sample.data["m0"], torch.Tensor)
            assert torch.min(sample.data["m0"]) >= 0
            assert torch.max(sample.data["m0"]) <= 1

            assert torch.min(sample.data["m1"]) >= 0
            assert torch.max(sample.data["m1"]) <= 1

            assert torch.min(sample.data["m2"]) >= 0
            assert torch.max(sample.data["m2"]) <= 1
            assert torch.min(sample.data["m3"]) >= 0
            assert torch.max(sample.data["m3"]) <= 1

            assert torch.min(sample.data["m4"]) >= 0
            assert torch.max(sample.data["m4"]) <= 1

            assert sample.data["m0"].size() == torch.Size([3, 28, 28])
            assert len(dataset) == 10000
            sample = dataset[:100]
            if input_dataset_test["missing_ratio"] > 0:
                assert hasattr(sample, "masks")
                assert torch.all(sample.masks["m0"] == 1)
                for m in ["m1", "m2", "m3", "m4"]:
                    assert torch.all(
                        sample.data[m][(1 - sample.masks[m].int()).bool()] == 0
                    )
                    assert not torch.all(sample.data[m][sample.masks[m]] == 0)

                    assert torch.all(sample.masks["m0"] == True)
                    assert not torch.all(sample.masks["m0"] == sample.masks["m1"])
        else:
            warnings.warn(
                message="PolyMNIST dataset is not downloaded at provided path and therefore"
                + "has not been tested"
            )



import logging
from typing import Dict, Union

import numpy as np
import torch
import torch.distributions as dist
from pythae.models.base.base_utils import ModelOutput
from pythae.models.nn.base_architectures import BaseDecoder, BaseEncoder
from pythae.models.normalizing_flows.base import BaseNF
from pythae.models.normalizing_flows.maf import MAF, MAFConfig
from torch.nn import ModuleDict

from multivae.models.nn.default_architectures import (
    BaseDictEncoders,
    MultipleHeadJointEncoder,
)

from ...data.datasets.base import MultimodalBaseDataset
from ..dcca import DCCA, DCCAConfig
from ..joint_models import BaseJointModel
from ..nn.default_architectures import BaseDictEncoders
from .jnf_dcca_config import JNFDccaConfig

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class JNFDcca(BaseJointModel):

    """
    The JNFDcca model.

    Args:

        model_config (JNFDccaConfig): An instance of JNFConfig in which any model's parameters is
            made available.

        encoders (Dict[str,BaseEncoder]): A dictionary containing the modalities names and the encoders for each
            modality. Each encoder is an instance of Pythae's BaseEncoder.

        decoders (Dict[str,BaseDecoder]): A dictionary containing the modalities names and the decoders for each
            modality. Each decoder is an instance of Pythae's BaseDecoder.

        joint_encoder (BaseEncoder) : An instance of BaseEncoder that takes all the modalities as an input.
            If none is provided, one is created from the unimodal encoders. Default : None.

        flows (Dict[str,BaseNF]) : A dictionary containing the modalities names and the flows to use for
            each modality. If None is provided, a default MAF flow is used for each modality.

        dcca_networks (Dict[str, BaseEncoder]) : A dictionary containing the networks to use in the DCCA module
            for each modality. If None is provided, default MLPs are used.

    """

    def __init__(
        self,
        model_config: JNFDccaConfig,
        encoders: Dict[str, BaseEncoder] = None,
        decoders: Dict[str, BaseDecoder] = None,
        joint_encoder: Union[BaseEncoder, None] = None,
        flows: Dict[str, BaseNF] = None,
        dcca_networks: Dict[str, BaseEncoder] = None,
        **kwargs,
    ):
        self.dcca_config = DCCAConfig(
            n_modalities=model_config.n_modalities,
            embedding_dim=model_config.embedding_dcca_dim,
            use_all_singular_values=model_config.use_all_singular_values,
        )

        if dcca_networks is None:
            if model_config.input_dims is None:
                raise AttributeError(
                    "Please provide dcca_networks or input_dims for the modalities in the model_config."
                    " The model can't be initialized if both are missing."
                )
            else:
                if len(model_config.input_dims.keys()) != model_config.n_modalities:
                    raise AttributeError(
                        f"The provided number of input_dims {len(model_config.input_dims.keys())} doesn't"
                        f"match the number of modalities ({model_config.n_modalities} in model config "
                    )
                dcca_networks = BaseDictEncoders(
                    model_config.input_dims, model_config.embedding_dcca_dim
                )
        else:
            model_config.custom_architectures.append("dcca_networks")

        super().__init__(model_config, encoders, decoders, joint_encoder, **kwargs)

        # The default encoders for this model have (embedding_dcca_dim, ) as input_size
        if encoders is None:
            encoders_input_dims = {
                k: (model_config.embedding_dcca_dim,) for k in dcca_networks
            }
            self.set_encoders(
                BaseDictEncoders(encoders_input_dims, model_config.latent_dim)
            )

        # The default joint_encoder for this model is engineered from the DCCA networks and
        # not from the encoders
        if joint_encoder is None:
            # Create a MultiHead Joint Encoder MLP
            self.set_joint_encoder(
                MultipleHeadJointEncoder(dcca_networks, model_config)
            )

        self.DCCA_module = DCCA(self.dcca_config, dcca_networks)
        self.dcca_networks = self.DCCA_module.networks

        if flows is None:
            flows = dict()
            for modality in self.encoders:
                flows[modality] = MAF(MAFConfig(input_dim=(model_config.latent_dim,)))
        else:
            model_config.custom_architectures.append("flows")

        self.set_flows(flows)
        self.model_name = "JNFDcca"
        self.warmup = model_config.warmup
        self.nb_epochs_dcca = model_config.nb_epochs_dcca
        self.reset_optimizer_epochs = [
            self.nb_epochs_dcca,
            self.nb_epochs_dcca + self.warmup,
        ]

    def set_flows(self, flows: Dict[str, BaseNF]):
        # check that the keys corresponds with the encoders keys
        if flows.keys() != self.encoders.keys():
            raise AttributeError(
                f"The keys of provided flows : {list(flows.keys())}"
                f" doesn't match the keys provided in encoders {list(self.encoders.keys())}"
                " or input_dims."
            )

        # Check that the flows are instances of BaseNF and that the input_dim for the
        # flows matches the latent_dimension

        self.flows = ModuleDict()
        for m in flows:
            if isinstance(flows[m], BaseNF):
                if flows[m].input_dim == (self.latent_dim,):
                    self.flows[m] = flows[m]
                else:
                    raise AttributeError(
                        "The provided flows don't have the right input dim."
                    )
            else:
                raise AttributeError(
                    "The provided flows must be instances of the Pythae's BaseNF "
                    " class."
                )
        return

    def _set_torch_no_grad_on_joint_vae(self):
        # After the warmup, we freeze the architecture of the joint encoder and decoders
        self.joint_encoder.requires_grad_(False)
        self.decoders.requires_grad_(False)

    def _set_torch_no_grad_on_dcca_module(self):
        self.DCCA_module.requires_grad_(False)

    def forward(self, inputs: MultimodalBaseDataset, **kwargs):
        epoch = kwargs.pop("epoch", 1)

        if epoch <= self.nb_epochs_dcca:
            return self.DCCA_module(inputs)
        else:
            self._set_torch_no_grad_on_dcca_module()

        # First compute the joint ELBO
        joint_output = self.joint_encoder(inputs.data)
        mu, log_var = joint_output.embedding, joint_output.log_covariance

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        z_joint = qz_xy.rsample()

        recon_loss = 0

        # Decode in each modality
        len_batch = 0
        for mod in self.decoders:
            x_mod = inputs.data[mod]
            len_batch = len(x_mod)
            recon_mod = self.decoders[mod](z_joint).reconstruction
            recon_loss += (
                -self.recon_log_probs[mod](recon_mod, x_mod) * self.rescale_factors[mod]
            ).sum()

        # Compute the KLD to the prior
        KLD = -0.5 * torch.sum(1 + log_var - mu.pow(2) - log_var.exp())

        if epoch <= self.warmup + self.nb_epochs_dcca:
            return ModelOutput(
                recon_loss=recon_loss / len_batch,
                KLD=KLD / len_batch,
                loss=(recon_loss + KLD) / len_batch,
                metrics=dict(kld_prior=KLD, recon_loss=recon_loss / len_batch, ljm=0),
            )

        else:
            self._set_torch_no_grad_on_joint_vae()
            ljm = 0
            for mod in self.encoders:
                dcca_embed = self.DCCA_module.networks[mod](inputs.data[mod]).embedding
                mod_output = self.encoders[mod](dcca_embed)
                mu0, log_var0 = mod_output.embedding, mod_output.log_covariance

                sigma0 = torch.exp(0.5 * log_var0)
                qz_x0 = dist.Normal(mu0, sigma0)

                # Compute -ln q_\phi_mod(z_joint|x_mod)
                flow_output = self.flows[mod](z_joint)
                z0 = flow_output.out

                ljm += -(
                    qz_x0.log_prob(z0).sum(dim=-1) + flow_output.log_abs_det_jac
                ).sum()

            return ModelOutput(
                recon_loss=recon_loss / len_batch,
                KLD=KLD / len_batch,
                loss=ljm / len_batch,
                ljm=ljm / len_batch,
                metrics=dict(
                    kld_prior=KLD,
                    recon_loss=recon_loss / len_batch,
                    ljm=ljm / len_batch,
                ),
            )

    def encode(
        self,
        inputs: MultimodalBaseDataset,
        cond_mod: Union[list, str] = "all",
        N: int = 1,
        **kwargs,
    ) -> ModelOutput:
        cond_mod = super().encode(inputs, cond_mod, N, **kwargs).cond_mod

        mcmc_steps = kwargs.pop("mcmc_steps", 100)
        n_lf = kwargs.pop("n_lf", 10)
        eps_lf = kwargs.pop("eps_lf", 0.01)
        return_mean = kwargs.pop("return_mean", False)
        if len(cond_mod) == self.n_modalities:
            output = self.joint_encoder(inputs.data)
            sample_shape = [] if N == 1 else [N]
            if return_mean:
                z = torch.stack([output.embedding] * N) if N > 1 else output.embedding
            else:
                z = dist.Normal(
                    output.embedding, torch.exp(0.5 * output.log_covariance)
                ).rsample(sample_shape)
            if N > 1 and kwargs.pop("flatten", False):
                N, l, d = z.shape
                z = z.reshape(l * N, d)
            return ModelOutput(z=z, one_latent_space=True)

        elif len(cond_mod) != 1:
            z = self.sample_from_poe_subset(
                cond_mod,
                inputs.data,
                ax=None,
                mcmc_steps=mcmc_steps,
                n_lf=n_lf,
                eps_lf=eps_lf,
                K=N,
                divide_prior=True,
            )
            if N > 1 and kwargs.pop("flatten", False):
                N, l, d = z.shape
                z = z.reshape(l * N, d)
            return ModelOutput(z=z, one_latent_space=True)

        elif len(cond_mod) == 1:
            cond_mod = cond_mod[0]
            dcca_embed = self.DCCA_module.networks[cond_mod](
                inputs.data[cond_mod]
            ).embedding
            output = self.encoders[cond_mod](dcca_embed)
            sample_shape = [] if N == 1 else [N]

            z0 = dist.Normal(
                output.embedding, torch.exp(0.5 * output.log_covariance)
            ).rsample(sample_shape)
            flow_output = self.flows[cond_mod].inverse(
                z0.reshape(-1, self.latent_dim)
            )  # The reshaping is because MAF flows doesn't handle
            # any shape of input data (*,*input_dim)
            z = flow_output.out.reshape(z0.shape)

            if N > 1 and kwargs.pop("flatten", False):
                z = z.reshape(-1, self.latent_dim)

            return ModelOutput(z=z, one_latent_space=True)
        else:
            raise AttributeError(
                f"Modality of name {cond_mod} not handled. The"
                f" modalities that can be encoded are {list(self.encoders.keys())}"
            )

    def sample_from_moe_subset(self, subset: list, data: dict):
        """Sample z from the mixture of posteriors from the subset.
        Torch no grad is activated, so that no gradient are computed durin the forward pass of the encoders.

        Args:
            subset (list): the modalities to condition on
            data (list): The data
            K (int) : the number of samples per datapoint
        """
        # Choose randomly one modality for each sample
        n_batch = len(data[list(data.keys())[0]])

        indices = np.random.choice(subset, size=n_batch)
        zs = torch.zeros((n_batch, self.latent_dim)).to(
            data[list(data.keys())[0]].device
        )

        for m in subset:
            with torch.no_grad():
                encoder_output = self.encoders[m](
                    self.DCCA_module.networks[m](data[m][indices == m]).embedding
                )
                mu, log_var = encoder_output.embedding, encoder_output.log_covariance
                zs[indices == m] = dist.Normal(mu, torch.exp(0.5 * log_var)).rsample()
        return zs

    def compute_poe_posterior(
        self, subset: list, z_: torch.Tensor, data: list, divide_prior=True, grad=True
    ):
        """Compute the log density of the product of experts for Hamiltonian sampling.

        Args:
            subset (list): the modalities of the poe posterior
            z_ (torch.Tensor): the latent variables (len(data[0]), latent_dim)
            data (list): _description_
            divide_prior (bool) : wether or not to divide by the prior

        Returns:
            tuple : likelihood and gradients
        """

        lnqzs = 0

        z = z_.clone().detach().requires_grad_(grad)

        if divide_prior:
            # print('Dividing by the prior')
            lnqzs += (0.5 * (torch.pow(z, 2) + np.log(2 * np.pi))).sum(dim=1)

        for m in subset:
            # Compute lnqz
            flow_output = self.flows[m](z)
            vae_output = self.encoders[m](
                self.DCCA_module.networks[m](data[m]).embedding
            )
            mu, log_var, z0 = (
                vae_output.embedding,
                vae_output.log_covariance,
                flow_output.out,
            )

            log_q_z0 = (
                -0.5
                * (
                    log_var
                    + np.log(2 * np.pi)
                    + torch.pow(z0 - mu, 2) / torch.exp(log_var)
                )
            ).sum(dim=1)
            lnqzs += log_q_z0 + flow_output.log_abs_det_jac  # n_data_points x 1

        if grad:
            g = torch.autograd.grad(lnqzs.sum(), z)[0]
            return lnqzs, g
        else:
            return lnqzs

    def sample_from_poe_subset(
        self,
        subset,
        data,
        ax=None,
        mcmc_steps=300,
        n_lf=10,
        eps_lf=0.01,
        K=1,
        divide_prior=True,
    ):
        """Sample from the product of experts using Hamiltonian sampling.

        Args:
            subset (List[int]):
            gen_mod (int):
            data (dict or MultimodalDataset):
            K (int, optional): . Defaults to 100.
        """
        logger.info(
            f"starting to sample from poe_subset, divide prior = {divide_prior}"
        )

        # Multiply the data to have multiple samples per datapoints
        n_data = len(data[list(data.keys())[0]])
        data = {d: torch.cat([data[d]] * K) for d in data}
        device = data[list(data.keys())[0]].device

        n_samples = len(data[list(data.keys())[0]])
        acc_nbr = torch.zeros(n_samples, 1).to(device)

        # First we need to sample an initial point from the mixture of experts
        z0 = self.sample_from_moe_subset(subset, data)
        z = z0

        # fig, ax = plt.subplots()
        pos = []
        grad = []
        for i in range(mcmc_steps):
            pos.append(z[0].detach().cpu())

            # print(i)
            gamma = torch.randn_like(z, device=device)
            rho = gamma  # / self.beta_zero_sqrt

            # Compute ln q(z|X_s)
            ln_q_zxs, g = self.compute_poe_posterior(
                subset, z, data, divide_prior=divide_prior
            )

            grad.append(g[0].detach().cpu())

            H0 = -ln_q_zxs + 0.5 * torch.norm(rho, dim=1) ** 2
            # print(H0)
            # print(model.G_inv(z).det())
            for k in range(n_lf):
                # z = z.clone().detach().requires_grad_(True)
                # log_det = G(z).det().log()

                # g = torch.zeros(n_samples, model.latent_dim).cuda()
                # for i in range(n_samples):
                #    g[0] = -grad(log_det, z)[0][0]

                # step 1
                rho_ = rho - (eps_lf / 2) * (-g)

                # step 2
                z = z + eps_lf * rho_

                # z_ = z_.clone().detach().requires_grad_(True)
                # log_det = 0.5 * G(z).det().log()
                # log_det = G(z_).det().log()

                # g = torch.zeros(n_samples, model.latent_dim).cuda()
                # for i in range(n_samples):
                #    g[0] = -grad(log_det, z_)[0][0]

                # Compute the updated gradient
                ln_q_zxs, g = self.compute_poe_posterior(subset, z, data, divide_prior)

                # print(g)
                # g = (Sigma_inv @ (z - mu).T).reshape(n_samples, 2)

                # step 3
                rho__ = rho_ - (eps_lf / 2) * (-g)

                # tempering
                beta_sqrt = 1

                rho = rho__
                # beta_sqrt_old = beta_sqrt

            H = -ln_q_zxs + 0.5 * torch.norm(rho, dim=1) ** 2
            # print(H, H0)

            alpha = torch.exp(H0 - H)
            # print(alpha)

            # print(-log_pi(best_model, z, best_model.G), 0.5 * torch.norm(rho, dim=1) ** 2)
            acc = torch.rand(n_samples).to(device)
            moves = (acc < alpha).type(torch.int).reshape(n_samples, 1)

            acc_nbr += moves

            z = z * moves + (1 - moves) * z0
            z0 = z

        pos = torch.stack(pos)
        grad = torch.stack(grad)
        if ax is not None:
            ax.plot(pos[:, 0], pos[:, 1])
            ax.quiver(pos[:, 0], pos[:, 1], grad[:, 0], grad[:, 1])

            # plt.savefig('monitor_hmc.png')
        # 1/0
        print(acc_nbr[:10] / mcmc_steps)
        sh = (n_data, self.latent_dim) if K == 1 else (K, n_data, self.latent_dim)
        z = z.detach().resize(*sh)
        return z.detach()



import numpy as np
import pytest
import torch
from encoders import Encoder_test, Encoder_test_multilatents

from multivae.data.datasets.base import IncompleteDataset, MultimodalBaseDataset
from multivae.models.base.base_config import BaseAEConfig
from multivae.models.mopoe.mopoe_config import MoPoEConfig
from multivae.models.mopoe.mopoe_model import MoPoE
from multivae.models.nn.default_architectures import Decoder_AE_MLP, ModelOutput
from multivae.samplers.gaussian_mixture import (
    GaussianMixtureSampler,
    GaussianMixtureSamplerConfig,
)
from multivae.samplers.iaf_sampler import IAFSampler, IAFSamplerConfig


@pytest.fixture
def dataset(request):
    # Create simple small dataset
    data = dict(
        mod1=torch.Tensor([[1.0, 2.0], [4.0, 5.0]]),
        mod2=torch.Tensor([[67.1, 2.3, 3.0], [1.3, 2.0, 3.0]]),
        mod3=torch.Tensor([[37, 2, 4, 1], [8, 9, 7, 0]]),
        mod4=torch.Tensor([[37, 2, 4, 1], [8, 9, 7, 0]]),
    )
    labels = np.array([0, 1, 0, 0])
    dataset = MultimodalBaseDataset(data, labels)

    return dataset


@pytest.fixture(params=["one_latent_space", "multi_latent_spaces"])
def archi_and_config(beta, request):
    if request.param == "one_latent_space":
        # Create an instance of mvae model
        config1 = BaseAEConfig(input_dim=(2,), latent_dim=5)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=5)
        config3 = BaseAEConfig(input_dim=(4,), latent_dim=5)

        encoders = dict(
            mod1=Encoder_test(config1),
            mod2=Encoder_test(config2),
            mod3=Encoder_test(config3),
            mod4=Encoder_test(config3),
        )

        model_config = MoPoEConfig(
            n_modalities=4,
            latent_dim=5,
            input_dims=dict(mod1=(2,), mod2=(3,), mod3=(4,), mod4=(4,)),
            beta=beta,
        )

        decoders = dict(
            mod1=Decoder_AE_MLP(config1),
            mod2=Decoder_AE_MLP(config2),
            mod3=Decoder_AE_MLP(config3),
            mod4=Decoder_AE_MLP(config3),
        )

    else:
        config1 = BaseAEConfig(input_dim=(2,), latent_dim=5, style_dim=4)
        config2 = BaseAEConfig(input_dim=(3,), latent_dim=5, style_dim=2)
        config3 = BaseAEConfig(input_dim=(4,), latent_dim=5, style_dim=3)

        encoders = dict(
            mod1=Encoder_test_multilatents(config1),
            mod2=Encoder_test_multilatents(config2),
            mod3=Encoder_test_multilatents(config3),
            mod4=Encoder_test_multilatents(config3),
        )
        model_config = MoPoEConfig(
            n_modalities=4,
            latent_dim=5,
            input_dims=dict(mod1=(2,), mod2=(3,), mod3=(4,), mod4=(4,)),
            beta=beta,
            use_modality_specific_spaces=True,
            modalities_specific_dim=dict(mod1=4, mod2=2, mod3=3, mod4=3),
        )
        decoders = dict(
            mod1=Decoder_AE_MLP(BaseAEConfig(input_dim=(2,), latent_dim=9)),
            mod2=Decoder_AE_MLP(BaseAEConfig(input_dim=(3,), latent_dim=7)),
            mod3=Decoder_AE_MLP(BaseAEConfig(input_dim=(4,), latent_dim=8)),
            mod4=Decoder_AE_MLP(BaseAEConfig(input_dim=(4,), latent_dim=8)),
        )

    return dict(encoders=encoders, decoders=decoders, model_config=model_config)


@pytest.fixture(params=[1.0, 1.5, 2.0])
def beta(request):
    beta = request.param

    return beta


@pytest.fixture(params=[True, False])
def model(archi_and_config, request):
    custom = request.param
    if custom:
        model = MoPoE(**archi_and_config)
    else:
        model = MoPoE(archi_and_config["model_config"])
    return model


class Test_gmm_sampler():


    @pytest.fixture(params = [4,10])
    def sampler_config(self,request):

        config = GaussianMixtureSamplerConfig(
            n_components=request.param
        )

    @pytest.fixture
    def sampler(self,sampler_config, model):

        sampler = GaussianMixtureSampler(model,sampler_config)
        return sampler

    def test_fit(self,sampler, dataset):

        sampler.fit(dataset)

        assert hasattr(sampler, 'gmm')
        assert sampler.is_fitted

        if sampler.model.multiple_latent_spaces:
            assert hasattr(sampler, 'gmm')

    def test_sample(self,sampler, dataset):

        sampler.fit(dataset)

        output = sampler.sample(100)

        assert isinstance(output, ModelOutput)
        assert hasattr(output, 'z')
        assert output.z.shape == (100,sampler.model.latent_dim)


class Test_iaf_sampler:
    @pytest.fixture(params=[0, 1])
    def sampler_config(self, request):
        if request.param == 0:
            return IAFSamplerConfig(n_made_blocks=3, n_hidden_in_made=4, hidden_size=64)
        else:
            return IAFSamplerConfig(
                n_made_blocks=1, n_hidden_in_made=1, hidden_size=512
            )

    @pytest.fixture
    def sampler(self, sampler_config, model):
        sampler = IAFSampler(model, sampler_config)
        return sampler

    def test_fit(self, sampler, dataset):
        sampler.fit(dataset)

        assert hasattr(sampler, "flows_models")
        assert hasattr(sampler, "iaf_models")

        assert sampler.is_fitted

        if sampler.model.multiple_latent_spaces:
            for m in sampler.model.encoders:
                assert m in sampler.iaf_models.keys()

        # test sample
        output = sampler.sample(100)

        assert isinstance(output, ModelOutput)
        assert hasattr(output, "z")
        assert output.z.shape == (100, sampler.model.latent_dim)

        if sampler.model.multiple_latent_spaces:
            assert hasattr(output, "modalities_z")



from typing import Literal

from torchvision import transforms
from torchvision.datasets import CelebA

from .base import DatasetOutput, MultimodalBaseDataset


class CelebAttr(MultimodalBaseDataset): # pragma: no cover
    def __init__(
        self,
        root: str,
        split: str,
        transform=None,
        target_transform=None,
        attributes: Literal["18", "40"] = "18",
        download=False,
    ):
        self.root = root

        if transform is None:
            transform = transforms.Compose(
                [
                    transforms.Resize(64),
                    transforms.CenterCrop(64),
                    transforms.ToTensor(),
                ]
            )
        self.transform = transform

        self.torchvision_dataset = CelebA(
            root=root,
            split=split,
            target_type="attr",
            transform=transform,
            download=download,
        )

        if attributes == "18":
            self.attributes_to_keep = [
                4,
                5,
                8,
                9,
                11,
                12,
                15,
                17,
                18,
                20,
                21,
                22,
                26,
                28,
                31,
                32,
                33,
                35,
            ]
        else:
            self.attributes_to_keep = range(40)

        self.attr_to_idx = {
            "Sideburns": 30,
            "Black_Hair": 8,
            "Wavy_Hair": 33,
            "Young": 39,
            "Heavy_Makeup": 18,
            "Blond_Hair": 9,
            "Attractive": 2,
            "5_o_Clock_Shadow": 0,
            "Wearing_Necktie": 38,
            "Blurry": 10,
            "Double_Chin": 14,
            "Brown_Hair": 11,
            "Mouth_Slightly_Open": 21,
            "Goatee": 16,
            "Bald": 4,
            "Pointy_Nose": 27,
            "Gray_Hair": 17,
            "Pale_Skin": 26,
            "Arched_Eyebrows": 1,
            "Wearing_Hat": 35,
            "Receding_Hairline": 28,
            "Straight_Hair": 32,
            "Big_Nose": 7,
            "Rosy_Cheeks": 29,
            "Oval_Face": 25,
            "Bangs": 5,
            "Male": 20,
            "Mustache": 22,
            "High_Cheekbones": 19,
            "No_Beard": 24,
            "Eyeglasses": 15,
            "Bags_Under_Eyes": 3,
            "Wearing_Necklace": 37,
            "Wearing_Lipstick": 36,
            "Big_Lips": 6,
            "Narrow_Eyes": 23,
            "Chubby": 13,
            "Smiling": 31,
            "Bushy_Eyebrows": 12,
            "Wearing_Earrings": 34,
        }

        self.idx_to_attr = {v: k for k, v in self.attr_to_idx.items()}

    def __getitem__(self, index):
        img, target = self.torchvision_dataset[index]

        return DatasetOutput(
            data=dict(image=img, attributes=target[self.attributes_to_keep].float()),
            labels=target[self.attributes_to_keep],
        )

    def __len__(self):
        return self.torchvision_dataset.__len__()



import logging
from typing import Union

import numpy as np
import torch
import torch.distributions as dist
from pythae.models.nn.base_architectures import BaseEncoder

from ...data import MultimodalBaseDataset
from ..base import BaseMultiVAE
from ..nn.default_architectures import MultipleHeadJointEncoder
from .joint_model_config import BaseJointModelConfig

logger = logging.getLogger(__name__)
console = logging.StreamHandler()
logger.addHandler(console)
logger.setLevel(logging.INFO)


class BaseJointModel(BaseMultiVAE):
    """
    Base Class for models using a joint encoder.

    Args:

        model_config (BaseJointModelConfig): The configuration of the model.

        encoders (Dict[BaseEncoder]): A dictionary containing the modalities names and the encoders for each
            modality. Each encoder is an instance of Pythae's BaseEncoder class.

        decoder (Dict[BaseDecoder]): A dictionary containing the modalities names and the decoders for each
            modality. Each decoder is an instance of Pythae's BaseDecoder class.

        joint_encoder (BaseEncoder) : An instance of BaseEncoder that takes all the modalities as an input.
            If none is provided, one is created from the unimodal encoders. Default : None.
    """

    def __init__(
        self,
        model_config: BaseJointModelConfig,
        encoders: dict = None,
        decoders: dict = None,
        joint_encoder: Union[BaseEncoder, None] = None,
        **kwargs,
    ):
        super().__init__(model_config, encoders, decoders)

        if joint_encoder is None:
            # Create a MultiHead Joint Encoder MLP
            joint_encoder = MultipleHeadJointEncoder(self.encoders, model_config)
        else:
            self.model_config.custom_architectures.append("joint_encoder")

        self.set_joint_encoder(joint_encoder)

    def set_joint_encoder(self, joint_encoder):
        "Checks that the provided joint encoder is an instance of BaseEncoder."

        if not issubclass(type(joint_encoder), BaseEncoder):
            raise AttributeError(
                (
                    f"The joint encoder must inherit from BaseEncoder class from "
                    "pythae.models.base_architectures.BaseEncoder. Refer to documentation."
                )
            )
        self.joint_encoder = joint_encoder

    def compute_joint_nll(
        self, inputs: MultimodalBaseDataset, K: int = 1000, batch_size_K: int = 100
    ):
        """
        Return the estimated negative log-likelihood summed over the input batch.
        The negative log-likelihood is estimated using importance sampling.

        Args:
            inputs : the data to compute the joint likelihood"""

        # First compute all the parameters of the joint posterior q(z|x,y)

        joint_output = self.joint_encoder(inputs.data)
        mu, log_var = joint_output.embedding, joint_output.log_covariance

        sigma = torch.exp(0.5 * log_var)
        qz_xy = dist.Normal(mu, sigma)
        # And sample from the posterior
        z_joint = qz_xy.rsample([K])  # shape K x n_data x latent_dim
        z_joint = z_joint.permute(1, 0, 2)
        n_data, _, latent_dim = z_joint.shape

        # Then iter on each datapoint to compute the iwae estimate of ln(p(x))
        ll = 0
        for i in range(n_data):
            start_idx = 0
            stop_idx = min(start_idx + batch_size_K, K)
            lnpxs = []
            while start_idx < stop_idx:
                latents = z_joint[i][start_idx:stop_idx]

                # Compute p(x_m|z) for z in latents and for each modality m
                lpx_zs = 0  # ln(p(x,y|z))
                for mod in inputs.data:
                    decoder = self.decoders[mod]
                    recon = decoder(latents)[
                        "reconstruction"
                    ]  # (batch_size_K, *decoder_output_shape)
                    x_m = inputs.data[mod][i]  # (*input_shape)
                    lpx_zs += (
                        self.recon_log_probs[mod](recon, x_m)
                        .reshape(recon.size(0), -1)
                        .sum(-1)
                    )

                # Compute ln(p(z))
                prior = dist.Normal(0, 1)
                lpz = prior.log_prob(latents).sum(dim=-1)

                # Compute posteriors -ln(q(z|x,y))
                qz_xy = dist.Normal(mu[i], sigma[i])
                lqz_xy = qz_xy.log_prob(latents).sum(dim=-1)

                ln_px = torch.logsumexp(lpx_zs + lpz - lqz_xy, dim=0)
                lnpxs.append(ln_px)

                # next batch
                start_idx += batch_size_K
                stop_idx = min(stop_idx + batch_size_K, K)

            ll += torch.logsumexp(torch.Tensor(lnpxs), dim=0) - np.log(K)

        return -ll



import os

import numpy as np
import pytest
from pythae.data.datasets import DatasetOutput
from torchvision.datasets import MNIST

from multivae.data.datasets.base import MultimodalBaseDataset
from multivae.data.datasets.mnist_svhn import MnistSvhn


@pytest.mark.slow
class Test:
    @pytest.fixture
    def input_dataset_test(self, tmpdir):
        data_path = os.path.join(tmpdir, "data")
        split = "test"

        return dict(data_path=data_path, split=split)

    def test_create_dataset(self, input_dataset_test):
        try:
            mnist = MNIST(
                input_dataset_test["data_path"],
                train=(input_dataset_test["split"] == "train"),
                download=False,
            )
        except:  # If the dataset is not available don't run the test
            return

        dataset = MnistSvhn(**input_dataset_test)
        assert isinstance(dataset, MultimodalBaseDataset)
        sample = dataset[0]
        assert type(sample) == DatasetOutput
        assert len(dataset) == 50000
